/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HH_
# define TOOL_OPTIONS_HH_

# include <vector>
# include <string>
# include <set>
# include <utility>

# include <wipal/tool/endianness.hh>

# define wp_opt_none		""
# define wp_opt_802cmp		"8"
# define wp_opt_attr		"a:"
# define wp_opt_bytescmp	"b"
# define wp_opt_colheads	"cC"
# define wp_opt_defaultcmp	"d"
# define wp_opt_errs		"eE"
# define wp_opt_help		"h"
# define wp_opt_frm_idx		"iI"
# define wp_opt_mapping		"m:"
# define wp_opt_endianness	"nN"
# define wp_opt_prism_noise	"pP"
# define wp_opt_quiet		"q"
# define wp_opt_ignore_ref	"r:"
# define wp_opt_timecmp		"t"
# define wp_opt_usec		"uU"
# define wp_opt_version		"v"

# define wp_help_make(L, M) "  -" #L "   " M "\n"

# define wp_help_802cmp							\
  wp_help_make(8, "Only compare 802.11 frames (no Prism, no PCAP).")

# define wp_help_attr							\
  wp_help_make(a, "Attributes for unique frames identification "	\
                  "(default: " wp_attr_default ").")

# define wp_help_bytescmp					\
  wp_help_make(b, "Only compare frame bytes (no PCAP header).")

# define wp_help_colheads						\
  wp_help_make(c, "Do not print column headers "			\
                  "(default if output is not a tty).")			\
  wp_help_make(C, "Print column headers (default if output is a tty).")

# define wp_help_defaultcmp						\
  wp_help_make(d, "Compare everything: PCAP data and bytes (default).")

# define wp_help_errs						\
  wp_help_make(e, "Do not print error values (default).")	\
  wp_help_make(E, "Print error values.")

# define wp_help_help				\
  wp_help_make(h, "Print this help.")

# define wp_help_frm_idx				\
  wp_help_make(i, "Do not print frame indices.")	\
  wp_help_make(I, "Print frame indices (default).")

# define wp_help_mapping				\
  wp_help_make(m, "MAC addresses mapping file "		\
	          "(default: " wp_mapping_default ").")

# define wp_help_endianness					\
  wp_help_make(n, "Force little endian for PHY headers.")	\
  wp_help_make(N, "Force big endian for PHY headers.")

# define wp_help_prism_noise						\
  wp_help_make(p, "Prism noise fields have no special meaning (default).") \
  wp_help_make(P, "Consider a non-zero Prism noise field as a PHY error.")

# define wp_help_quiet							\
  wp_help_make(q, "Quiet ; produce minimal output (no column headers, etc.).")

# define wp_help_ignore_ref \
  wp_help_make(r, "Ignore reference.  Blacklist a given x-y reference frame.")

# define wp_help_timecmp						\
wp_help_make(t, "Assume Prism headers and compare 802.11 frames, host time, " \
		"and MAC time. Use a precision of 106us.")

# define wp_help_usec						\
  wp_help_make(u, "Do not print microseconds (default).")	\
  wp_help_make(U, "Print microseconds.")

# define wp_help_version			\
  wp_help_make(v, "Print version information.")

# define wp_attr_tmp			"timestamp"
# define wp_attr_dst_tmp		"dest-timestamp"
# define wp_attr_src_tmp		"source-timestamp"
# define wp_attr_bss_tmp		"bssid-timestamp"
# define wp_attr_src_bss_tmp		"source-bssid-timestamp"
# define wp_attr_seq_tmp		"seqctl-timestamp"
# define wp_attr_seq_bss_tmp		"seqctl-bssid-timestamp"
# define wp_attr_seq_dst_bss_tmp	"seqctl-dest-bssid-timestamp"
# define wp_attr_seq_src_bss_tmp	"seqctl-source-bssid-timestamp"

# define wp_attr_default		wp_attr_seq_bss_tmp
# define wp_mapping_default		"mapping"

namespace tool
{

  namespace opt
  {

    struct options
    {
      typedef std::vector<std::string>			input_type;
      typedef std::vector<std::string>			output_type;
      typedef std::set< std::pair<unsigned, unsigned> >	ref_blacklist_type;

      enum
	{
	  no_input			= 0,
	  no_output			= 0,
	  arbitrary_number_of_inputs	= 8,
	  arbitrary_number_of_outputs	= 8,
	};

      enum comp_method
	{
	  cmp_default,
	  cmp_802,
	  cmp_bytes,
	  cmp_time
	};

      input_type		input;
      output_type		output;
      std::string		mapping;
      std::string		attributes;
      endian::endianness	endianness;
      ref_blacklist_type	ref_blacklist;
      comp_method		cmp;
      bool			prism_noise_means_error;
      bool			print_column_headers;
      bool			print_errors;
      bool			print_frame_indices;
      bool			print_microseconds;
      bool			print_summary;
    };

    options	parse_args(int argc, char* const* argv,
			   const std::string& available_options,
			   int ninputs, int noutputs);

  } // End of namespace tool::opt.

} // End of namespace tool.

# include "options.hxx"

#endif // TOOL_OPTIONS_HH_
