/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_TIME_ADJUSTER_HXX_
# define WIFI_FRAME_FILTER_TIME_ADJUSTER_HXX_

# include "time_adjuster.hh"

# include <boost/shared_ptr.hpp>
# include <boost/shared_array.hpp>

# include <wipal/phy/prism_header.hh>
# include <wipal/wifi/frame/filter/microseconds_stamper.hh>
# include <wipal/wifi/frame/filter/linear_regression_synchronizer.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class I1, class I2, class HT, class B1, class B2>
	time_adjuster_iterator<I1, I2, HT, B1, B2>::
	time_adjuster_iterator(const iterable_type& i, bool end):
	  super_type (),
	  iterable_ (&i),
	  next1_ (end ? i.last1_ : i.first1_),
	  next2_ (end ? i.last2_ : i.first2_)
	{
	  if (not end)
	    increment();
	}

	template <class I1, class I2, class HT, class B1, class B2>
	bool
	time_adjuster_iterator<I1, I2, HT, B1, B2>::
	equal(const time_adjuster_iterator& rhs) const
	{
	  if (not rhs.value())
	    return not this->value();
	  return next1_ == rhs.next1_;
	}

	template <class I1, class I2, class HT, class B1, class B2>
	void
	time_adjuster_iterator<I1, I2, HT, B1, B2>::increment()
	{
	  using tool::endian::endianness;

	  typedef tool::microseconds				microseconds;
	  typedef typename time_converter<>::time_values	time_values;
	  typedef pcapxx::pkthdr				pkthdr;
	  typedef boost::shared_ptr<pkthdr>			pkthdr_ptr;
	  typedef boost::shared_array<uint8_t>			bytes_ptr;


	  if (next1_ == iterable_->last1_)
	    {
	      this->value() = boost::none_t ();

	      // Flush next2_ so it may print a summary of its operation.
	      while (next2_ != iterable_->last2_)
		++next2_;

	      return;
	    }
	  update_coefs();

	  const tool::microseconds	us (next1_->microseconds() *
					    coefs_->first + coefs_->second);
	  const time_values		tv = iterable_->convert_(us);

	  const pcapxx::pkthdr*	pcaph	= next1_->pcap_header().get();
	  const size_t		caplen	= pcaph->caplen;
	  const bool		swapped	= need_swap(iterable_->phy_end_,
						    pcaph->swapped);
	  const pkthdr_ptr	v_pcap_header (new pkthdr (*pcaph));
	  const bytes_ptr	v_bytes       (new uint8_t[caplen]);

	  v_pcap_header->ts = tv.pcaptime;
	  memcpy(v_bytes.get(), next1_->bytes().get(), caplen);
	  internals::adjust(reinterpret_cast<HT*> (v_bytes.get()),
			    caplen, swapped, tv);
	  this->value() =
	    value_type (us, pcapxx::frame_descriptor (v_pcap_header, v_bytes));
	  ++next1_;
	}

	template <class I1, class I2, class HT, class B1, class B2>
	void
	time_adjuster_iterator<I1, I2, HT, B1, B2>::update_coefs()
	{
	  const tool::microseconds	us = next1_->microseconds();
	  const I2&			last2 = iterable_->last2_;

	  // FIXME: There used to be a bug in the following check.  It checked
	  // FIXME: for next2_->*second*.microseconds() <= us instead of
	  // FIXME: next2_->first.microseconds().  Unfortunately the current
	  // FIXME: test suite does not catch this bug, though this is a
	  // FIXME: dramatic malfunction.  One should write a regression
	  // FIXME: test for this case.
	  while (next2_ != last2 and next2_->first.microseconds() <= us)
	    {
	      coefs_ = next2_->coefs;
	      ++next2_;
	    }
	  if (not coefs_)
	    {
	      if (next2_ != last2)
		coefs_ = next2_->coefs;
	      else
		throw std::runtime_error ("No reference frames: "
					  "impossible to synchronize");
	    }
	}

	template <>
	inline
	time_converter<>
	time_reference<prism::header>(const pcapxx::pkthdr&	pcap_header,
				      const prism::header*	bytes,
				      tool::endian::endianness	phy_end)
	{
	  using tool::endian::need_swap;

	  if (pcap_header.caplen < sizeof (prism::header))
	    return time_converter<> (pcap_header.ts, 0, 0,
				     tool::microseconds (0));

	  const bool	 swapped  = need_swap(phy_end, pcap_header.swapped);
	  const uint32_t hosttime = bytes->hosttime.get(swapped);
	  const uint32_t mactime  = bytes->mactime.get(swapped);

	  return time_converter<> (pcap_header.ts, hosttime, mactime,
				   tool::microseconds (mactime));
	}

	template <>
	inline
	void
	adjust<prism::header>(prism::header* bytes, size_t caplen, bool swap,
			      const time_converter<>::time_values& time)
	{
	  if (caplen < sizeof (prism::header))
	    return;

	  bytes->hosttime.set(time.hosttime, swap);
	  bytes->mactime.set(time.mactime, swap);
	}

      } // End of namespace wifi::frame::filter::internals.


      template <class I1, class I2, class HT, class B>
      time_adjuster<I1, I2, HT, B>::
      time_adjuster(const I1& first1, const I1& last1,
		    const I2& first2, const I2& last2,
		    const pcapxx::frame_descriptor&	tref,
		    tool::endian::endianness		phy_end):
	convert_ (internals::time_reference(*tref.pcap_header(),
					    reinterpret_cast<const HT*>
					                  (tref.bytes().get()),
					    phy_end)),
	phy_end_ (phy_end),
	first1_ (first1),
	last1_ (last1),
	first2_ (first2),
	last2_ (last2)
      {
      }

      namespace internals
      {

	template <class HT, class I1, class I2, class F>
	struct provide_time_adjuster
	{
	  provide_time_adjuster(const I1& first1, const I1& last1,
				const I2& first2, const I2& last2,
				F&				func,
				bool				filter_prism,
				tool::endian::endianness	phy_end):
	    first1_ (first1),
	    last1_ (last1),
	    first2_ (first2),
	    last2_ (last2),
	    func_ (func),
	    filter_prism_ (filter_prism),
	    phy_end_ (phy_end)
	  {
	  }

	  template <class Synchronizer>
	  void
	  operator () (const Synchronizer& s)
	  {
	    if (first2_ == last2_)
	      throw std::runtime_error ("No packets in reference trace. "
					"Cannot synchronize with no time "
					"reference and no unique frames");
	    if (filter_prism_)
	      {
		typedef non_noisy_prism<I2>			nnp;
		typedef typename nnp::const_iterator		n_iterator;
		typedef microseconds_stamper<n_iterator, HT>	us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename Synchronizer::const_iterator	s_iterator;
		typedef time_adjuster<u_iterator, s_iterator, HT> adjuster;

		nnp		n (first1_, last1_);
		us_stamper	u (n.begin(), n.end(), phy_end_);
		adjuster	a (u.begin(), u.end(),
				   s.begin(), s.end(), *first2_, phy_end_);
		func_(a);
	      }
	    else
	      {
		typedef microseconds_stamper<I2, HT>		us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename Synchronizer::const_iterator	s_iterator;
		typedef time_adjuster<u_iterator, s_iterator, HT> adjuster;

		us_stamper	u (first1_, last1_, phy_end_);
		adjuster	a (u.begin(), u.end(),
				   s.begin(), s.end(), *first2_, phy_end_);
		func_(a);
	      }
	  }

	private:
	  const I1&			first1_;
	  const I1&			last1_;
	  const I2&			first2_;
	  const I2&			last2_;
	  F&				func_;
	  bool				filter_prism_;
	  tool::endian::endianness	phy_end_;
	};

      } // End of namespace wifi::frame::filter::internals.

      template <class U, class HT, template <class, class, class> class Int,
		class I1, class I2, class F, class BL>
      void
      provide_time_adjuster(const I1& first1, const I1&	last1,
			    const I2& first2, const I2&	last2,
			    addr_mapping&		mapping,
			    F&				func,
			    bool			filter_prism,
			    tool::endian::endianness	phy_end,
			    const BL&			blist)
      {
	internals::provide_time_adjuster<HT, I1, I2, F> func2 (first1, last1,
							       first2, last2,
							       func,
							       filter_prism,
							       phy_end);

	provide_lr_synchronizer<U, HT, Int>(first1, last1,
					    first2, last2,
					    mapping, func2,
					    filter_prism, phy_end, blist);
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_TIME_ADJUSTER_HXX_
